/* Hello World Example for UMD HPC Clusters */
/* Copyright @2020-2021 University of Maryland */
/* This software is provided "AS IS", without warranty of any kind, express or 
implied, including but not limited to the warranties of merchantability, fitness 
for a particular purpose and noninfringement. In no event shall the authors or 
copyright holders be liable for any claim, damages or other liability, whether 
in an action of contract, tort or otherwise, arising from, out of or in connection 
with the software or the use or other dealings in the software.

   You are allowed to copy, modify, and distribute this code as desired
   provided you leave in the above disclaimer.
*/
/* This code has the following compilation options depending on the 
   definition of preprocessor macros:

   MPI_BUILD:
	if defined, do an MPI build
	otherwise, do a non-MPI build

   HELLOUMD_COMPILER;
	this is a string giving the compiler version.
	Should be a module tag (e.g. gcc/8.2.0, etc)

   HELLOUMD_MPI
	For MPI builds, similar to HELLOUMD_COMPILER but 
	with information re MPI
*/
  
	
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h> /* for getenv */
#include <sys/types.h> /* for getpid */
#include <string.h>

#define DEFAULT_MESSAGE "Hello UMD"

#ifdef MPI_BUILD
	/* MPI builds need mpi.h */
#	include <mpi.h>
#endif

/* We always use OpenMP */
/* OpenMP header */
#include <omp.h>

#define MAXBUF 1000
#define VERSION "1.5"

/* These macros allow us to use HELLOUMD_COMPILER, HELLOUMD_MPI in strings */
#define STRINGIFY2( x) #x
#define STRINGIFY(x) STRINGIFY2(x)

/* Store the HELLOUMD_COMPILER value in named string, or default if missing */
void get_compiler_string(char *outbuf)
{
#ifdef HELLOUMD_COMPILER
	strncpy(outbuf, STRINGIFY(HELLOUMD_COMPILER), MAXBUF);
#else
	strncpy(outbuf, "<unknown compiler>", MAXBUF);
#endif
	outbuf[MAXBUF+1] = '\0';
}


/* Store the HELLOUMD_MPI value in named string, or default if missing */
void get_mpi_string(char *outbuf)
{
#ifdef HELLOUMD_MPI
	strncpy(outbuf, STRINGIFY(HELLOUMD_MPI), MAXBUF);
#else
	strncpy(outbuf, "<unknown MPI library>", MAXBUF);
#endif
	outbuf[MAXBUF+1] = '\0';
}

void print_id_string()
{	char compstr[MAXBUF+1];
  	char mpistr[MAXBUF+1];
	fprintf(stderr, "hello-umd: Version %s\n", VERSION);
  	get_compiler_string(compstr);
	fprintf(stderr, "Built for compiler: %s\n", compstr);
#	ifdef MPI_BUILD
	get_mpi_string(mpistr);
	fprintf(stderr, "with MPI support( usgin MPI library %s)\n", mpistr);
#	endif
}

void usage(char *progname)
{	
	fprintf(stderr, "%s: Simple MPI/OpenMP hello world code\n", 
		progname);
	print_id_string();
	fprintf(stderr, "\nUsage:\n");
	fprintf(stderr, "%s [ -h ] [ -t MAXTHREADS ] [-m MESSAGE]\n\n", 
		progname);
	fprintf(stderr, "Prints a hello message from each task/thread.\n");
	fprintf(stderr, "The -h flag prints this help text and exits.\n\n");
	fprintf(stderr, "If MESSAGE is given, that is the message that will\n");
	fprintf(stderr, "be printed (with 'from thread A of B ...' appended)\n");
	fprintf(stderr, "The default MESSAGE is '%s'\n\n",DEFAULT_MESSAGE);


	fprintf(stderr, "This code supports multiple threads using OpenMP\n");
	fprintf(stderr, "The maximum number of threads to be used can be\n");
	fprintf(stderr, "using the -t MAXTHREADS flag.  If that is not set,\n");
	fprintf(stderr, "the value of the env var OMP_NUM_THREADS will be\n");
	fprintf(stderr, "used, and if -t is not set and OMP_NUM_THREADS is\n");
	fprintf(stderr, "set, then it defaults to 1 thread.\n");
	fprintf(stderr, "If MAXTHREADS is set to 0, then the code will limit\n");
	fprintf(stderr, "the number of threads based on OMP_NUM_THREADS if set\n");
	fprintf(stderr, "or the number of CPU cores available on the system.\n");
#	ifdef MPI_BUILD
	fprintf(stderr, "\nThis version is built with MPI support; run with\n");
	fprintf(stderr, "mpirun or similar to use multiple MPI tasks.\n");
	fprintf(stderr, "\nThis software is copyright 2020-2021 by the\n");
	fprintf(stderr, "University of Maryland.\n");
#	endif
}

int main (argc, argv)
     int argc;
     char *argv[];
{
  /* We default rank => 0 and numranks=> 1 for non-MPI jobs */
  int rank=0;
  int numranks=1;
  /* We default thread => 0 and maxthreads => 1 for non OpenMP jobs */
  int thread=0;
  int maxthreads=1;
  int tmp;
  int opt;
  pid_t pid;
#ifdef MPI_BUILD
  /* Because we do minimal MPI, use the least amount of parallelism supported*/
  int required_mpithreading = MPI_THREAD_FUNNELED; 
  int provided_mpithreading;
#endif
  char myhostname[MAXBUF + 1];
  char compstr[MAXBUF+1];
  char mpistr[MAXBUF+1];
  char *tmpstr;
  char *message = DEFAULT_MESSAGE;

  /* Set up compiler and MPI identification strings */
  get_compiler_string(compstr);
  get_mpi_string(mpistr);

  tmpstr = getenv("OMP_NUM_THREADS");
  if ( tmpstr )
  {	/* OMP_NUM_THREADS is defined */
	maxthreads = atoi(tmpstr);
  }

  while ( (opt = getopt(argc, argv, "ht:m:")) != -1)
  {	switch (opt)
	{  case 'h':
		usage(argv[0]);
		exit(0);
		break;
	   case 't':
		maxthreads = atoi(optarg);
		break;
	   case 'm':
		message = optarg;
		break;
	   default:
		usage(argv[0]);
		fprintf(stderr, "\nError processing command options.\n");
		exit(EXIT_FAILURE);
		break;
	}
   }

#ifdef MPI_BUILD
   if ( maxthreads == 1 )
   {	/* Not doing any OpenMP threading, drop required_mpithreading to
	   MPI_THREAD_SINGLE */
	required_mpithreading = MPI_THREAD_SINGLE;
   }
#endif

   /* Set the maximum number of threads for this process, unless 
     maxthreads is set to zero, in which case let OpenMP defaults take over.
   */
   if ( maxthreads )
   {  	omp_set_num_threads(maxthreads);
   	tmp = omp_get_max_threads();
	if ( tmp > maxthreads )
	{   fprintf(stderr, "omp_get_max_threads returned %d > maxthreads=%d\n",
			tmp, maxthreads);
	    fprintf(stderr, "Aborting\n");
	    exit(EXIT_FAILURE);
	} 
	if ( tmp < maxthreads )
	{   fprintf(stderr, "omp_get_max_threads returned %d < maxthreads=%d\n",
		tmp, maxthreads);
	    fprintf(stderr, "Replacing maxthreads with %d\n", tmp);
	    maxthreads = tmp;
   	}
   }

#ifdef MPI_BUILD
  /* MPI builds require us to initialize MPI and get real values for rank and numranks */
  MPI_Init_thread(&argc, &argv, 
	required_mpithreading, &provided_mpithreading);	/* starts MPI */
  if ( provided_mpithreading < required_mpithreading )
  {	fprintf(stderr, "ERROR in MPI_Init_thread: unable to initialize\n");
  	fprintf(stderr, "requested threading support (%d), only able to\n",
		required_mpithreading);
	fprintf(stderr, "provided=%d\n", provided_mpithreading);
	exit(EXIT_FAILURE);
  }
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);	/* get current process id */
  MPI_Comm_size (MPI_COMM_WORLD, &numranks);	/* get number of ranks */
#endif

  /* Only print our program identification strings for first rank 
	of the job.  We are not OpenMP parallelized here, so not printed
	for all threads */
  if ( rank == 0  )
  {	print_id_string();
  }

  /* Get the host we are running on */
  gethostname(myhostname, MAXBUF);

 /* Print out the task identification string */
 #pragma omp parallel private(thread, pid)
 {	thread = omp_get_thread_num();
	pid = getpid();
  	printf( "%s from thread %d of %d, task %d of %d (pid=%d on host %s\n",
		message, thread, maxthreads, rank, numranks, pid, myhostname);
 }

#ifdef MPI_BUILD
  /* MPI Builds require some cleanup */
  MPI_Finalize();
#endif

  /* Done */
  return 0;
}
